SLASH_MYPOCKET1 = "/mp";
CurrentList = nil;

MyPocketSettings = {
	["ColorAll"] = "OFF",
	["EnableMinimap"] = "ON",
	["OpenWithBank"] = "OFF",
	["CloseWithBank"] = "ON",
	["ReturnEquiped"] = "OFF",
	["StackItems"] = "ON",
	["CompleteStackToBank"] = "OFF",
	["CompleteStackFromBank"] = "OFF",
	["LockOn"] = "ON", -- not a general option
	["MessageMissingInBag"] = "ON",
}

local myPocketBankOpen;
local isMovingSet = false;
local skipHandler = false;
local itemQueue;
local lastMovedBagFrom;
local lastMovedSlotFrom;
local pickedUpBag;
local pickedUpSlot;

isLocked = {};
for i=-1,11 do
	isLocked[i] = {};
end

function MyPocket_HandleEvent(self, event, arg1, arg2)
	
	if event == "ITEM_LOCK_CHANGED" then

		pickedUpBag = arg1;
		pickedUpSlot = arg2;

		--print(arg1, arg2, tostring(isLocked[arg1][arg2] or "not locked"))

		if arg1 and arg2 and isLocked[arg1][arg2] and not GetCursorInfo() then
			-- if item locked -> unlock and process 
			--print("unlocking",arg1,arg2)
			isLocked[arg1][arg2] = nil;
			
			--print("skip",skipHandler,lastMovedBagFrom,lastMovedSlotFrom)
			if itemQueue and not skipHandler and lastMovedBagFrom == arg1 and lastMovedSlotFrom == arg2 then
				WorkOnItemQueueByHandler();
			end
		end

	elseif event == "BAG_UPDATE" and not isMovingSet then
		ColorMissingItems();
		
	elseif event == "BANKFRAME_OPENED" then
		myPocketBankOpen = 1;
		
		if MyPocketSettings["OpenWithBank"] == "ON" then
			MyPocket:Show();
		end
		
	elseif event == "BANKFRAME_CLOSED" then
		myPocketBankOpen = nil;
		
		if MyPocketSettings["CloseWithBank"] == "ON" then
			MyPocket:Hide();
		end
		
	elseif event == "ADDON_LOADED" and arg1 == "MyPocket" then
		if MyPocketSets ~= nil then
			-- saved sets found
			if not CurrentList or not MyPocketSets[CurrentList] then
				-- no previous set set
				for setName, list in pairs(MyPocketSets) do
					--MyPocketStoredItemLinks = list;
					LoadPocketElements(list);
					CurrentList = setName;
					break;
				end
			else
				-- previous set found
				LoadPocketElements(MyPocketSets[CurrentList]);
			end
		else
			MyPocketSets = {};
			MyPocketSets["default"] = {};
			CurrentList = "default";
		end
		
		--init drop down
		UIDropDownMenu_SetText(MyPocketSelectButton, CurrentList);

		MyPocketCreateInterfaceOptions();
		MyPocketMiniMapButton.Init();
		ApplyInterfaceSettings();
	end
end

function PocketItemSlot_OnRightClick(frame)

	local frameName = frame:GetName();
	local item_info = MyPocketSets[CurrentList][frameName];
	local currentItems = MyPocketSets[CurrentList];

	if item_info and item_info.count then

		if item_info.count <= 1 then
			-- remove item
			PocketItemSlot_Remove(frame);
		else
			item_info.count = item_info.count - 1;
			frame.item_info = item_info;
			currentItems[frameName] = item_info;
			MyPocketSets[CurrentList] = currentItems;

			-- update icon / create frame
			SetIconAndColorAndText(frame);

			-- update colors
			ColorMissingItems();
		end
	end
end

function ToggleLockState(frame)
	if frame.item_info.isLocked then
		frame.item_info.isLocked = false;
	else
		frame.item_info.isLocked = true;
	end
end

function PocketItemSlot_ReceiveDrag(frame)
	local currentItems = MyPocketSets[CurrentList];
	local type, id, link = GetCursorInfo();
	local item_info;
	local frameName = frame:GetName();
	local itemCount;

	if type == "item" then
		-- add item to slot or increaase count
		
		if  pickedUpBag and pickedUpSlot then
			-- drag from bags/bank
			_, itemCount, _, _, _, _, _ = GetContainerItemInfo(pickedUpBag, pickedUpSlot);
		else
			-- drag from equipment
			itemCount = 1;
		end

		local _, _, _, _, _, _, _, stack, _, _, _, _, _, _, _, _, _ = GetItemInfo(link);

		if currentItems[frameName] and stack > 1 then
			item_info = currentItems[frameName];

			if item_info.link == link and item_info.count < stack then
				item_info.count = item_info.count + 1;
			else
				item_info.link = link;
				item_info.count = itemCount;
			end
		else
			item_info = {};
			item_info.link = link;
			item_info.count = itemCount;
		end
	elseif IsShiftKeyDown() and currentItems[frameName] then
		item_info = currentItems[frameName];
		MyPocketSettings["LockOn"] = "ON";
		ToggleLockState(frame);
	else
		-- increase count on leftclick with nothing on cursor 
		item_info = currentItems[frameName];

		if item_info and item_info.count then
			local _, _, _, _, _, _, _, stack, _, _, _, _, _, _, _, _, _ = GetItemInfo(item_info.link);

			if item_info.count < stack then
				item_info.count = item_info.count + 1;
			end
		else
			-- return on empty slot
			return;
		end
	end

	frame.item_info = item_info;
	currentItems[frameName] = item_info;
	MyPocketSets[CurrentList] = currentItems;

	SetIconAndColorAndText(frame);
	ColorMissingItems();
end

function SetIconFrame(frame) 
	frame.item_icon = frame:CreateTexture("Icon_" ..  frame:GetName() , "BACKGROUND", nil, 7);
	--frame.item_icon:SetFrameLevel(10);
	frame.item_icon:SetWidth(34);
	frame.item_icon:SetHeight(34);
	frame.item_icon:SetPoint("TOPLEFT", 0, 0);

	if not frame.item_icon:IsVisible() then
		frame.item_icon:Show();
	end
end

function MyPocket_OnLoad(frame) 
	local MyPocketsBagFrame;
	local NewMyPocketsBagFrame;

	for r=1,7 do
		for c=1,7 do

			if r == 1 and c == 1 then
				MyPocketsBagFrame = CreateFrame("Button", "PocketElementR1C1", frame, "PocketElement");
				MyPocketsBagFrame:SetPoint("TOPLEFT", "NameNewPocketInput", "BOTTOMLEFT", -5, -5);
			elseif r ~= 1 and c == 1 then
				MyPocketsBagFrame = CreateFrame("Button", "PocketElementR" .. r .. "C1", frame, "PocketElement");
				MyPocketsBagFrame:SetPoint("TOPLEFT", "PocketElementR" .. (r-1) .. "C1", "BOTTOMLEFT", 0, 0);
			else
				NewMyPocketsBagFrame = CreateFrame("Button", "PocketElementR" .. r .. "C" .. c, frame, "PocketElement");
				NewMyPocketsBagFrame:SetPoint("TOPLEFT", MyPocketsBagFrame, "TOPRIGHT", 0, 0);
				MyPocketsBagFrame = NewMyPocketsBagFrame;
			end		

			--frame.item_icon:SetTexture("Interface/Buttons/UI-EmptySlot");
			MyPocketsBagFrame:SetBackdrop({ 
				bgFile = "Interface/Buttons/UI-Slot-Background", 
				--edgeFile = "edgeFile", tile = false, tileEdge = true, tileSize = 0, edgeSize = 32, 
				insets = { left = 0, right = -18, top = 1, bottom = -17},
				--backdropBorderColor = { r=0.7, g=1, b=0.7, a=1 },
				backdropColor = {a=1, r=0.1, g=0.1, b=0.1 },
			});

			SetIconFrame(MyPocketsBagFrame);
		end
	end
end

function MyPocketCreateInterfaceOptions()
	-- general settings tab

	MyPocketSettingsPanel = {};
	MyPocketSettingsPanel.panel = CreateFrame( "Frame", "MyPocketInterfaceSettings", UIParent );
	MyPocketSettingsPanel.panel.name = "MyPocket";
	InterfaceOptions_AddCategory(MyPocketSettingsPanel.panel);

	-- check boxes

	local button = CreateFrame("CheckButton","MyPocketSettingColorAll",MyPocketSettingsPanel.panel,"ChatConfigCheckButtonTemplate")
	button:SetFrameStrata("HIGH");
	button:SetScript("OnClick", EvaluateCheckbox);
	button:SetPoint("TOPLEFT", 40, -40);
	button.setting = "ColorAll";
	button.tooltip = "Normally it subtracts your inventory items from the requested ones and colors the missing amount. Check to disable";
	getglobal(button:GetName() .. 'Text'):SetText("Color all missing items");
	if MyPocketSettings["ColorAll"] == "ON" then button:SetChecked(true) end

	button = CreateFrame("CheckButton","MyPocketSettingEnableMinimap", button ,"ChatConfigCheckButtonTemplate")
	button:SetFrameStrata("HIGH");
	button:SetScript("OnClick", EvaluateCheckbox);
	button:SetPoint("TOPLEFT", 0, -20);
	button.setting = "EnableMinimap";
	button.tooltip = "Check to show minimap icon";
	getglobal(button:GetName() .. 'Text'):SetText("Enable minimap icon");
	if MyPocketSettings["EnableMinimap"] == "ON" then button:SetChecked(true) end

	button = CreateFrame("CheckButton","MyPocketSettingOpenWithBank", button ,"ChatConfigCheckButtonTemplate")
	button:SetFrameStrata("HIGH");
	button:SetScript("OnClick", EvaluateCheckbox);
	button:SetPoint("TOPLEFT", 0, -20);
	button.setting = "OpenWithBank";
	button.tooltip = "Check to open MyPocket with your bank";
	getglobal(button:GetName() .. 'Text'):SetText("Open with bank");
	if MyPocketSettings["OpenWithBank"] == "ON" then button:SetChecked(true) end

	button = CreateFrame("CheckButton","MyPocketSettingCloseWithBank", button ,"ChatConfigCheckButtonTemplate")
	button:SetFrameStrata("HIGH");
	button:SetScript("OnClick", EvaluateCheckbox);
	button:SetPoint("TOPLEFT", 0, -20);
	button.setting = "CloseWithBank";
	button.tooltip = "Check to close MyPocket with your bank";
	getglobal(button:GetName() .. 'Text'):SetText("Close with bank");
	if MyPocketSettings["CloseWithBank"] == "ON" then button:SetChecked(true) end

	button = CreateFrame("CheckButton","MyPocketSettingReturnEquiped", button ,"ChatConfigCheckButtonTemplate")
	button:SetFrameStrata("HIGH");
	button:SetScript("OnClick", EvaluateCheckbox);
	button:SetPoint("TOPLEFT", 0, -20);
	button.setting = "ReturnEquiped";
	button.tooltip = "Check to return equiped items to bank";
	getglobal(button:GetName() .. 'Text'):SetText("Return equiped items");
	if MyPocketSettings["ReturnEquiped"] == "ON" then button:SetChecked(true) end

	button = CreateFrame("CheckButton","MyPocketSettingMoveCompleteToBank", button ,"ChatConfigCheckButtonTemplate")
	button:SetFrameStrata("HIGH");
	button:SetScript("OnClick", EvaluateCheckbox);
	button:SetPoint("TOPLEFT", 0, -20);
	button.setting = "CompleteStackToBank";
	button.tooltip = "Item stacks are getting transferred as a whole to bank";
	getglobal(button:GetName() .. 'Text'):SetText("Don't split items when returning to bank");
	if MyPocketSettings["CompleteStackToBank"] == "ON" then button:SetChecked(true) end

	button = CreateFrame("CheckButton","MyPocketSettingMoveCompleteToBag", button ,"ChatConfigCheckButtonTemplate")
	button:SetFrameStrata("HIGH");
	button:SetScript("OnClick", EvaluateCheckbox);
	button:SetPoint("TOPLEFT", 0, -20);
	button.setting = "CompleteStackFromBank";
	button.tooltip = "Item stacks are getting transferred as a whole to bags";
	getglobal(button:GetName() .. 'Text'):SetText("Don't split items when picking up from bank");
	if MyPocketSettings["CompleteStackFromBank"] == "ON" then button:SetChecked(true) end

	button = CreateFrame("CheckButton","MyPocketSettingMessageMissingInBag", button ,"ChatConfigCheckButtonTemplate")
	button:SetFrameStrata("HIGH");
	button:SetScript("OnClick", EvaluateCheckbox);
	button:SetPoint("TOPLEFT", 0, -20);
	button.setting = "MessageMissingInBag";
	button.tooltip = "Create a chat message for missing items after picking something up from bank";
	getglobal(button:GetName() .. 'Text'):SetText("Create a chat message for missing items");
	if MyPocketSettings["MessageMissingInBag"] == "ON" then button:SetChecked(true) end

end

function EvaluateCheckbox(frame)
	if frame:GetChecked() and frame.setting then
		MyPocketSettings[frame.setting] = "ON";
	elseif not frame:GetChecked() and frame.setting then
		MyPocketSettings[frame.setting] = "OFF";
	end

	ApplyInterfaceSettings();
end

function ApplyInterfaceSettings()
	MyPocketMiniMapButton.UpdateVisibility();
	SetToggleButtonColor();
	ColorMissingItems();
end
	
function MyPocketSelectButton_Initialize()
	if MyPocketSets then
		for set_name, list in pairs(MyPocketSets) do
			local info = UIDropDownMenu_CreateInfo();
			info.text = set_name;
			--info.value = "OptionVariable";
			info.func = MyPocketSelectionMade;
			--info.func(self, info.arg1, info.arg2, checked)
			--info.notCheckable = true;
			info.arg1 = set_name;
			info.checked = (CurrentList and (CurrentList == set_name));
			UIDropDownMenu_AddButton(info);
		end
	end

	if CurrentList then
		UIDropDownMenu_SetText(MyPocketSelectButton, CurrentList);
	else
		UIDropDownMenu_SetText(MyPocketSelectButton, "default");
	end

	--[[
	local txt = UIDropDownMenu_GetText(MyPocketSelectButton);
	if not txt or txt == "" then
		UIDropDownMenu_SetText(MyPocketSelectButton, "default");
	end 
	]] 
end

function MyPocketSelectionMade(info, arg1, arg2, checked)
	CurrentList = arg1;
	UIDropDownMenu_SetText(MyPocketSelectButton, arg1);
	CloseDropDownMenus();
	LoadPocketSet(arg1);
end

function MyPocketSelectButton_OnClick() 
	ToggleDropDownMenu(1, nil, MyPocketSelectButton, MyPocketSelectButton, 8, 7);
end

function SlashCmdList.MYPOCKET(msg, editbox) 
	if MyPocket:IsVisible() then
		MyPocket:Hide();
	else
		MyPocket:Show();
	end
end

function SetIconAndColorAndText(frame)

	if frame.item_info and frame.item_info.link then
		local _, _, itemRarity, _, _, _, _, stack, _, _, _, _, _, _, _, _, _ = GetItemInfo(frame.item_info.link);

		local r, g, b, _ = GetItemQualityColor(itemRarity or 1);
		frame:GetNormalTexture():SetVertexColor(r, g, b);

		-- set icon texture
		frame.item_icon:SetTexture(GetItemIcon(frame.item_info.link));
		
		if not frame.item_icon:IsVisible() then
			frame.item_icon:Show();
		end

		local count_frame = frame.count_frame or frame:CreateFontString(frame:GetName() .. "ItemCountText", nil, "NumberFontNormalSmall");

		if stack and stack > 1 then
			count_frame:SetText(frame.item_info.count);
		else
			count_frame:SetText("");
		end

		count_frame:SetPoint("BOTTOMRIGHT",0,3);
		--frame:SetText(stack_frame);
		frame.count_frame = count_frame;

	end

end

function SetToggleButtonColor()
	if MyPocketSettings["LockOn"] == "ON" then
		
		PocketToggleLockButton:SetText("|cFF00FF00Lock On");
	else
		PocketToggleLockButton:SetText("|cFFFFD300Lock Off");
	end
end

function LoadPocketElements(pocketElements)
	if pocketElements then
		for _, child in pairs({ MyPocket:GetChildren() }) do
			if pocketElements[child:GetName()] then
				local item_info = pocketElements[child:GetName()];
				child.item_info = item_info;
				SetIconAndColorAndText(child);
			else
				if child.item_icon then
					child.item_icon:Hide();
					child:GetNormalTexture():SetVertexColor(1, 1, 1);
				end
				if child.count_frame then
					child.count_frame:SetText("");
				end
			end
		end
	end

	UIDropDownMenu_SetText(MyPocketSelectButton, CurrentList);
end

function LoadPocketSet(setName) 

	if setName and MyPocketSets[setName] then
		LoadPocketElements(MyPocketSets[setName]);
	end	

	ColorMissingItems();
end

function PocketItemSlot_OnEnter(frame)
	if frame.item_info and frame.item_icon:IsVisible() then
		GameTooltip_SetDefaultAnchor(GameTooltip, UIParent);
		GameTooltip:SetHyperlink(frame.item_info.link);
	end
end

function PocketItemSlot_OnLeave(frame)
	if frame.item_info and frame.item_icon:IsVisible() then
		GameTooltip:SetHyperlink(frame.item_info.link);
	end
end

function PocketItemSlot_Remove(frame)
	-- remove element from list

	if GameTooltip:IsVisible() and frame.item_info then
		GameTooltip:SetHyperlink(frame.item_info.link);	
		frame.item_info = nil;
	end

	if frame.item_icon then
		frame.item_icon:Hide();
	end

	if frame.count_frame then
		frame.count_frame:SetText("");
	end
	
	frame:GetNormalTexture():SetVertexColor(1, 1, 1);

	local currentItems = MyPocketSets[CurrentList];

	local frameName = frame:GetName();
	for listFrameName, _ in pairs(currentItems) do
		if listFrameName == frameName then
			currentItems[frameName] = nil;
		end
	end

	MyPocketSets[CurrentList] = currentItems;

	ColorMissingItems();
end

-- and locked items
function BuildPocketItemList()
	local pocketItemList = {};
	local pocketLockedItemList = {};
	local currentItems = MyPocketSets[CurrentList];

	if currentItems then
		for _, item_info in pairs(currentItems) do
		
			if pocketItemList[item_info.link] then
				pocketItemList[item_info.link] = pocketItemList[item_info.link] + item_info.count;
			else
				pocketItemList[item_info.link] = item_info.count;
			end

			-- locked items
			if item_info.isLocked and MyPocketSettings["LockOn"] == "ON" then
				if pocketLockedItemList[item_info.link] then
					pocketLockedItemList[item_info.link] = pocketLockedItemList[item_info.link] + item_info.count;
				else
					pocketLockedItemList[item_info.link] = item_info.count;
				end
			end
		end
	end

	return pocketItemList, pocketLockedItemList;
end

function BuildMissingItemList()
	local pocketMissingItemList = BuildPocketItemList();

	for bag = BACKPACK_CONTAINER, NUM_BAG_SLOTS do
		for slot = 1, GetContainerNumSlots(bag) do
			item_link = GetContainerItemLink(bag, slot);
			_, itemCount, _, _, _, _, _, _, _, _ = GetContainerItemInfo(bag, slot);	

			local requestedCount = pocketMissingItemList[item_link];
			
			if requestedCount then
				-- subtract items in pocket from requested ones
				if requestedCount - itemCount <= 0 then
					pocketMissingItemList[item_link] = nil;
				else
					pocketMissingItemList[item_link] = requestedCount - itemCount;
				end
			end
		end
	end

	-- remove equiped items
	for itemLink, count in pairs(pocketMissingItemList) do
		if IsEquippedItem(itemLink) then
			if count <= 1 then
				pocketMissingItemList[itemLink] = nil;
			else
				pocketMissingItemList[itemLink] = count - 1;
			end
		end
	end

	return pocketMissingItemList;
end

function ColorMissingItems()
	local pocketMissingItemList = BuildMissingItemList();
	local currentItems = MyPocketSets[CurrentList];

	for _, child in pairs({ MyPocket:GetChildren() }) do
		if currentItems and currentItems[child:GetName()] then
			local item_info = currentItems[child:GetName()];
			
			if pocketMissingItemList[item_info.link] then
				-- red border + normal icon

				child:GetNormalTexture():SetVertexColor(1, 0, 0);
				
				if item_info.isLocked and MyPocketSettings["LockOn"] == "ON" then
					child.item_icon:SetVertexColor(1, 1, 0);
					if child.count_frame then child.count_frame:SetVertexColor(1,1,0); end
				else
					child.item_icon:SetVertexColor(1, 1, 1);
					if child.count_frame then child.count_frame:SetVertexColor(1,1,1); end
				end

				if MyPocketSettings["ColorAll"] == "OFF" then
					-- dont mark all items red. Only the right amount
					local count = pocketMissingItemList[item_info.link] - item_info.count;				
					if count <= 0 then
						pocketMissingItemList[item_info.link] = nil;
					else
						pocketMissingItemList[item_info.link] = count; 
					end
				end
			elseif item_info.isLocked and MyPocketSettings["LockOn"] == "ON" then
				-- yellow border + icon
				child:GetNormalTexture():SetVertexColor(1, 1, 0);
				child.item_icon:SetVertexColor(1, 1, 0);
				if child.count_frame then child.count_frame:SetVertexColor(1,1,0); end
			else
				-- item rarity border + normal icon
				local _, _, itemRarity, _, _, _, _, stack, _, _, _, _, _, _, _, _, _ = GetItemInfo(item_info.link);
				local r, g, b, _ = GetItemQualityColor(itemRarity or 1);
				child:GetNormalTexture():SetVertexColor(r, g, b);
				child.item_icon:SetVertexColor(1, 1, 1);	
				if child.count_frame then child.count_frame:SetVertexColor(1,1,1); end
			end
		end
	end
end

function PocketLoadButton_OnClick()
	ClearCursor();
	isMovingSet = true;
	WorkOnItemQueue(false);
	isMovingSet = false;
end

function PocketReturnButton_OnClick()
	ClearCursor();
	isMovingSet = true;
	WorkOnItemQueue(true);
	isMovingSet = false;
end

function PocketToggleLockButton_OnClick()
	if MyPocketSettings["LockOn"] == "ON" then
		MyPocketSettings["LockOn"] = "OFF";
	else
		MyPocketSettings["LockOn"] = "ON";
	end

	-- recolor all
	ColorMissingItems();
	-- recolor Button
	SetToggleButtonColor();
end

function PocketNewSelectionButton_OnClick()
	local newListName = NameNewPocketInput:GetText();
	NameNewPocketInput:SetText("");
	NameNewPocketInput:ClearFocus();

	-- dont allow duplicates
	if not MyPocketSets[newListName] and newListName and newListName ~= "" then
		MyPocketSets[newListName] = {};
		CurrentList = newListName;
		LoadPocketSet(CurrentList);
	end
end

function PocketDeleteSelectionButton_OnClick()
	local listToDeleteName = UIDropDownMenu_GetText(MyPocketSelectButton);
	NameNewPocketInput:ClearFocus();
	
	-- delete
	if MyPocketSets[listToDeleteName] then
		MyPocketSets[listToDeleteName] = nil;
	end
	
	-- search for first remaining list
	local firstExistingList;
	for listName, list in pairs(MyPocketSets) do
		firstExistingList = listName;
		break;
	end

	-- recreate default if no other list present
	if firstExistingList then
		CurrentList = firstExistingList;
	else
		MyPocketSets["default"] = {};
		CurrentList = "default";
	end

	LoadPocketSet(CurrentList);
end

function MyPocketInfo_OnEnter(frame)
	local tooltip = _G["GameTooltip"];
    tooltip:SetOwner(frame, "ANCHOR_LEFT", 11, 5);--(frame:GetWidth() * 0.5), 5);
	tooltip:AddLine("MyPocket", 0, 1, 0);
	tooltip:AddLine("|cFFFFD300Left Click |cFFFFFFFFIncrease amount");
	tooltip:AddLine("|cFFFFD300Right Click |cFFFFFFFFDecrease amount");
	tooltip:AddLine("|cFFFFD300Middle Click |cFFFFFFFFRemove item");
	tooltip:AddLine("|cFFFFD300Shift Left Click |cFFFFFFFFUn-/Lock item");
	tooltip:AddLine("----------------------------------");
	tooltip:AddLine("|cFF00FF00Lock On |cFFFFFFFFLocking feature active");
	tooltip:AddLine("|cFFFFD300Lock Off |cFFFFFFFFLocking feature paused");
	tooltip:AddLine("|cFFFFFFFFLocked items stay in your pocked");
    tooltip:Show();
end

function MyPocketInfo_OnLeave(frame)
	local tooltip = _G["GameTooltip"];-- or MyPocketInfoTooltip;
	tooltip:Hide();
end

function ReduceRequestedAmount(itemList, itemLink, itemCountPickedUp)
	local newItemCount = itemList[itemLink] - itemCountPickedUp;
	if newItemCount <= 0 then
		itemList[itemLink] = nil;
	else
		itemList[itemLink] = newItemCount;
	end
end

function FindFreeAndIncompleteSlots(IsBank)
	local bagIDs;
	if IsBank then bagIDs = { -1,5,6,7,8,9,10 };
	else bagIDs = { 0,1,2,3,4 };
	end

	local freeBags = {};
	local bagType;

	if myPocketBankOpen then
		for _, bagID in pairs(bagIDs) do
			if (IsBank and bagID ~= -1) or (not IsBank and bagID ~= 0)  then			
				local invID = ContainerIDToInventoryID(bagID);
				local bagLink = GetInventoryItemLink("player",invID);
				
				if bagLink then 
					local itemType, itemSubType, _, _, _, _, classID, subclassID = select(6, GetItemInfo(bagLink))
					bagType = itemSubType;
				end
			else
				if IsBank then bagType = "Bank";
				else bagType = "Backpack";
				end
			end
			
			local slotsOfType = freeBags[bagType] or {}
			
			for slot=1, GetContainerNumSlots(bagID) do
				local _, itemCount, _, _, _, _, itemLink = GetContainerItemInfo(bagID, slot);
				local itemName, itemLink, _, _, _, _, _, itemStackCount, _, _, _ = GetItemInfo(itemLink or 0);
				local freeOfStackable = (itemStackCount or 0) - (itemCount or 0);

				if itemLink and freeOfStackable > 0 then
					-- something in pocket but stack is not full

					local incompleteBagOfType = slotsOfType[itemLink] or {};
					local incompleteSlot = {};
					
					incompleteSlot.bagID = bagID;
					incompleteSlot.slot = slot;
					incompleteSlot.free = freeOfStackable;
					tinsert(incompleteBagOfType, incompleteSlot);
					slotsOfType[itemLink] = incompleteBagOfType;

				elseif not itemLink then
					-- empty bag

					local emptyBagOfType = slotsOfType["empty"] or {};
					local emptySlot = {};
					
					emptySlot.bagID = bagID;
					emptySlot.slot = slot;
					--emptySlot.free = 1;
					tinsert(emptyBagOfType, emptySlot);
					slotsOfType["empty"] = emptyBagOfType;

				else
					-- full slot
				end
			end
			
			freeBags[bagType] = slotsOfType;
		end
	end

	return freeBags;
end

function DeterminePreferredBag(itemID, bagTopology, to_bank)
	if itemID then
		local itemType, itemSubType, _, _, _, _, classID, subclassID = select(6, GetItemInfo(itemID));

		local bagType;

		--if itemType == "Tradeskill" and (bagTopology["Enchanting Bag"] or bagTopology["Herb Bag"]) then
		--	if itemSubType == "Enchanting" then
		--		bagType = "Enchanting Bag";
		--	elseif itemSubType == "Herb" then
		--		bagType = "Herb Bag";
		--	end
		--if itemType == "Projectile" and bagTopology["Quiver"] then
		--	if itemSubType == "Arrow" then
		--		bagType =  "Quiver";
		--	elseif itemSubType == "Bullet" then
		--		bagType = "Ammo Pouch";
		--	elseif bagTopology["Bag"] then
		--		bagType = "Bag";
		--	else
		--		bagType = "noSpace";
		--	end
		--elseif itemType == "Recipe" and itemSubType == "Enchanting" and bagTopology["Enchanting Bag"] then
		--	bagType = "Enchanting Bag";
		if itemID == 6265 and bagTopology["Soul Bag"] then
			bagType = "Soul Bag";
		elseif not to_bank and bagTopology["Backpack"] then
			bagType = "Backpack";
		elseif to_bank and bagTopology["Bank"] then
			bagType = "Bank";
		elseif bagTopology["Bag"] then
			bagType = "Bag";
		else
			--bagType = "noSpace";
		end

		return bagType;
	end
end

function DetermineCorrectBagSlot(slotsWithCapacity, itemLink, preferredBagOfType, isStackable, toBank)
	local secondGuess;
	if toBank then secondGuess = "Bank"
	else secondGuess = "Backpack" end


	local theBagType;
	local theSubType;
	local containsIncompleteStack = false;
	local doStack = isStackable and MyPocketSettings["StackItems"] == "ON";

	if doStack and slotsWithCapacity[preferredBagOfType][itemLink] then
		-- get first available stackable slot
		theBagType = preferredBagOfType;
		theSubType = itemLink;
		containsIncompleteStack = true;
	elseif doStack and slotsWithCapacity[secondGuess][itemLink] then
		--search stackable in bank/backpack
		theBagType = secondGuess;
		theSubType = itemLink;
		containsIncompleteStack = true;
	elseif doStack and slotsWithCapacity["Bag"][itemLink] then
		--search stackable in bags
		theBagType = "Bag";
		theSubType = itemLink;
		containsIncompleteStack = true;
	elseif slotsWithCapacity[preferredBagOfType]["empty"] then
		-- put item in preferred bag in empty space
		theBagType = preferredBagOfType;
		theSubType = "empty";
	elseif slotsWithCapacity[secondGuess]["empty"] then
		-- search free bank/backpack
		theBagType = secondGuess;
		theSubType = "empty";
	elseif slotsWithCapacity["Bag"]["empty"] then
		-- search free bag
		theBagType = "Bag";
		theSubType = "empty";
	else

		if MyPocketSettings["StackItems"] == "OFF" then
			PrintMsg("Not enough space for " .. itemLink .. " (no stacking)");
		else
			PrintMsg("Not enough space for " .. itemLink);
		end
		return;
	end

	--print("for",itemLink,"decided",theBagType,theSubType)

	for id, firstBag in pairs(slotsWithCapacity[theBagType][theSubType]) do
		--print("fitting bag found", firstBag.bagID, firstBag.slot)
		return firstBag, id, containsIncompleteStack, theBagType, theSubType;
	end

	--print("found no fitting bag")
end

function WorkOnItemQueueByHandler()

	-- called by handler

	-- move first item in list
	-- remove item from list
	-- wait until item lock says --> is unlocked
	-- call function again from event handler

	--print("called by handler")

	for nr, item in pairs(itemQueue) do
		
		--print("now moving", item.item, item.toBagId, item.toSlot),
		isLocked[item.fromBagId][item.fromSlot] = true;

		if item.split then
			--print("splitting")
			lastMovedBagFrom = item.fromBagId;
			lastMovedSlotFrom = item.fromSlot;
			SplitContainerItem(item.fromBagId, item.fromSlot, item.amount);
		else
			--print("didnt split")
			isLocked[item.toBagId][item.toSlot] = true;
			lastMovedBagFrom = item.toBagId;
			lastMovedSlotFrom = item.toSlot;
			PickupContainerItem(item.fromBagId, item.fromSlot);
		end

		PickupContainerItem(item.toBagId, item.toSlot);
		itemQueue[nr] = nil;
		return;
	end

	itemQueue = nil;
end

function WorkOnItemQueue(toBank)
	itemQueue = CreateItemQueue(toBank);
	
	-- skip event handeler (ITEM_LOCK_CHANGED)
	skipHandler = true;

	for i=-1,11 do
		isLocked[i] = {};
	end

	-- return equiped items here
	for nr, item in pairs(itemQueue) do
		if item.isEquippedItem then
			PickupInventoryItem(item.fromSlot);
			itemQueue[nr] = nil;
			PickupContainerItem(item.toBagId, item.toSlot);
		end
	end

	-- move not stackable items
	for nr, item in pairs(itemQueue) do
		
		if not item.canSplit then
			--print("returning non stackable", item.item, item.toBagId, item.toSlot)
			PickupContainerItem(item.fromBagId, item.fromSlot);
			itemQueue[nr] = nil;
			PickupContainerItem(item.toBagId, item.toSlot);
		end
	end

	-- move stakable but only once
	for nr, item in pairs(itemQueue) do
		if not isLocked[item.toBagId][item.toSlot] and not isLocked[item.fromBagId][item.fromSlot] then

			isLocked[item.fromBagId][item.fromSlot] = true;
			isLocked[item.toBagId][item.toSlot] = true;
			
			if item.split then
				--print("splitting")
				lastMovedBagFrom = item.fromBagId;
				lastMovedSlotFrom = item.fromSlot;
				SplitContainerItem(item.fromBagId, item.fromSlot, item.amount);
			else
				--print("didnt split")
				lastMovedBagFrom = item.toBagId;
				lastMovedSlotFrom = item.toSlot;
				PickupContainerItem(item.fromBagId, item.fromSlot);
			end

			itemQueue[nr] = nil;
			PickupContainerItem(item.toBagId, item.toSlot);
		end
	end

	-- after last loop -> event should not have triggered -> set false -> event comes -> continue
	skipHandler = false;
end

function CreateEquipedItemsQueue(itemQueue, itemsToMove, slotsWithCapacity, itemsLocked)
	for slot=1,19 do
		local itemLink = GetInventoryItemLink("player", slot);

		if itemLink and itemsToMove[itemLink] and not itemsLocked[itemLink] then
			
			local preferredBagOfType = DeterminePreferredBag(itemLink, slotsWithCapacity ,true);
			local newBag, usedBagSubId, containsIncompleteStack, theBagType, theSubType = DetermineCorrectBagSlot(slotsWithCapacity, itemLink, preferredBagOfType, false, true);

			tinsert(itemQueue, {
				item = itemLink,
				toBagId = newBag.bagID, 
				toSlot = newBag.slot,
				--fromBagId = bagID,
				fromSlot = slot,
				amount = itemCount,
				split = false, -- maybe improve here
				canSplit = false,
				isEquippedItem = true} );	

			ReduceRequestedAmount(itemsToMove, itemLink, 1);
			slotsWithCapacity[theBagType][theSubType][usedBagSubId] = nil;
			RemoveEmptySubList(slotsWithCapacity[theBagType], theSubType);
			
		else
			--reduce locked amount
			if itemsLocked[itemLink] then
				local newLockCount = itemsLocked[itemLink] - 1;
				if newLockCount > 0 then
					itemsLocked[itemLink] = newLockCount;
				else
					itemsLocked[itemLink] = nil;
				end
			end
		end
	end
end

function CreateItemQueue(toBank)
	local bagIDs;
	local itemsToMove;
	local itemsLocked = {};
	local slotsWithCapacity = FindFreeAndIncompleteSlots(toBank);
	local itemQueue = {};
	local moveAsWhole = (toBank and MyPocketSettings["CompleteStackToBank"] == "ON") or
						(not toBank and MyPocketSettings["CompleteStackFromBank"] == "ON");

	if toBank then 
		bagIDs = { 0,1,2,3,4 };
		itemsToMove, itemsLocked = BuildPocketItemList();
	else 
		bagIDs = { -1,5,6,7,8,9,10 };
		itemsToMove = BuildMissingItemList();
	end

	
	if toBank and MyPocketSettings["ReturnEquiped"] == "ON" then
		CreateEquipedItemsQueue(itemQueue, itemsToMove, slotsWithCapacity, itemsLocked);
	end

	if myPocketBankOpen then
		for _, bagID in pairs(bagIDs) do
			for slot=1, GetContainerNumSlots(bagID) do
				
				local itemLink = GetContainerItemLink(bagID, slot) 
				local _, itemCount, _, _, _, _, _, _, _, _ = GetContainerItemInfo(bagID, slot);
				local actualPickedUp;

				while true do
				
					local itemCountToPickup = (itemsToMove[itemLink] or 0) - (itemsLocked[itemLink] or 0);

					if itemCountToPickup > 0 and itemCount > 0 then
						
						local maxStackCount = select(8, GetItemInfo(itemLink));
						local isStackable = maxStackCount > 1;
						local preferredBagOfType = DeterminePreferredBag(itemLink, slotsWithCapacity ,toBank);
						local newBag, usedBagSubId, containsIncompleteStack, theBagType, theSubType = DetermineCorrectBagSlot(slotsWithCapacity, itemLink, preferredBagOfType, isStackable, toBank);
						
						--print(itemLink, itemCountToPickup, preferredBagOfType, itemCount,  newBag.bagID, newBag.slot )

						if newBag and newBag.bagID and newBag.slot then						
							if containsIncompleteStack and newBag.free < itemCount then
								-- item count in source slot is bigger than free space on target stack
								-- reduce to free space in bag
								itemCountToPickup = newBag.free;	
							end

							-- found requested item
							if itemCount <= itemCountToPickup or moveAsWhole then
								-- pick whole stack/item			
								tinsert(itemQueue, {
													item = itemLink,
													toBagId = newBag.bagID, 
													toSlot = newBag.slot,
													fromBagId = bagID,
													fromSlot = slot,
													amount = itemCount,
													split = false,
													canSplit = isStackable,
													isEquippedItem = false} );	
								actualPickedUp = itemCount;
							else
								-- split stack
								tinsert(itemQueue, {
													item = itemLink,
													toBagId = newBag.bagID,
													toSlot = newBag.slot,
													fromBagId = bagID,
													fromSlot = slot,
													amount = itemCountToPickup,
													split = true,
													canSplit = isStackable,
													isEquippedItem = false} );
								actualPickedUp = itemCountToPickup;
							end

							ReduceRequestedAmount(itemsToMove, itemLink, actualPickedUp);

							--print("1:",containsIncompleteStack, newBag.free - actualPickedUp > 0 )
							--print("2:", isStackable , not containsIncompleteStack , maxStackCount - actualPickedUp > 0)

							if containsIncompleteStack and newBag.free - actualPickedUp > 0 then
								--print("free",slotsWithCapacity[theBagType][theSubType][usedBagSubId], "reducing to",newBag.free - actualPickedUp)
								-- there is still space on the stack after putting it down
								
								local updateBag = slotsWithCapacity[theBagType][theSubType][usedBagSubId]
								updateBag.free = newBag.free - actualPickedUp;
								slotsWithCapacity[theBagType][theSubType][usedBagSubId] = updateBag;
								break;
							
							elseif isStackable and not containsIncompleteStack and maxStackCount - actualPickedUp > 0 then
								-- bag is an empty bag
								-- moving incomplete stack there
								-- --> move bag from type empty to itemLink

								--print("empty to stackable")

								local movedBag = slotsWithCapacity[theBagType][theSubType][usedBagSubId]
								slotsWithCapacity[theBagType][theSubType][usedBagSubId] = nil;
								RemoveEmptySubList(slotsWithCapacity[theBagType], theSubType);
								
								movedBag.free = maxStackCount - actualPickedUp;

								if not slotsWithCapacity[theBagType][itemLink] then
									slotsWithCapacity[theBagType][itemLink] = {}
								end

								tinsert(slotsWithCapacity[theBagType][itemLink], movedBag)
								break;
							
							else
								-- no space left on this stack
								--print("removing");
								itemCount = itemCount - actualPickedUp;
								slotsWithCapacity[theBagType][theSubType][usedBagSubId] = nil;
								RemoveEmptySubList(slotsWithCapacity[theBagType], theSubType);
							end
						else
							-- nothing free for item/s
							break;
						end
					else
						-- nothing requested or item count on slot = 0
						-- or item locked -> reduce locked amount 
						if itemsLocked[itemLink] then
							local newLockCount = itemsLocked[itemLink] - itemCount;
							if newLockCount > 0 then
								itemsLocked[itemLink] = newLockCount;
							else
								itemsLocked[itemLink] = nil;
							end
						end
						break;
					end
				end
			end
		end
	end

	-- message for still missing items after transfer
	if not toBank and itemsToMove and MyPocketSettings["MessageMissingInBag"] == "ON" then
		local stillMissing = "Missing in inventory:";
		for link, amount in pairs(itemsToMove) do
			stillMissing = stillMissing .. "\n" .. amount .. " " .. link;
		end

		-- nothing missing
		if stillMissing ~= "Missing in inventory:" then 
			print(stillMissing);
		end
	end

	return itemQueue;
end

function RemoveEmptySubList(list, key)
	for _,_ in pairs(list[key]) do
		return;
	end
	list[key] = nil;
end

function PrintMsg(msg)
	if msg then
		DEFAULT_CHAT_FRAME:AddMessage("|cFFEB33FFMyPocket: |cFFFFFFFF"..msg)
	end
end